# -*- coding: utf-8 -*-
"""
sceptre.connection_manager

This module implements a ConnectionManager class, which simplifies and manages
Boto3 calls.
"""

import functools
import logging
import os
import random
import threading
import time
import warnings
from typing import Optional, Dict, Tuple, Any

import boto3
import deprecation
from botocore.credentials import Credentials
from botocore.exceptions import ClientError

from sceptre.exceptions import InvalidAWSCredentialsError, RetryLimitExceededError
from sceptre.helpers import mask_key, create_deprecated_alias_property


def _retry_boto_call(func):
    """
    Retries a Boto3 call up to 30 times if request rate limits are hit.

    Between each try we wait a random amount with a max of that time being 45 seconds.
    Specifically we are picking number between a ceiling (delay_cap) of 45 seconds and the last
    delay multiplied by 2.5, rounding to two decimal places.  You can read more
    here: https://aws.amazon.com/blogs/architecture/exponential-backoff-and-jitter/
    If rate limits are hit 30 times, _retry_boto_call raises a
    sceptre.exceptions.RetryLimitExceededException.

    :param func: A function that uses boto calls
    :type func: function
    :returns: The decorated function.
    :rtype: function
    :raises: sceptre.exceptions.RetryLimitExceededException
    """
    logger = logging.getLogger(__name__)

    @functools.wraps(func)
    def decorated(*args, **kwargs):
        max_retries = 30
        attempts = 1
        mdelay = 1
        delay_cap = 45
        while attempts < max_retries:
            try:
                return func(*args, **kwargs)
            except ClientError as e:
                if e.response["Error"]["Code"] == "Throttling":
                    logger.error("Request limit exceeded, pausing {}...".format(mdelay))
                    time.sleep(mdelay)

                    # Using De-correlated Jitter Algorithm
                    # We are picking number between a ceiling (delay_cap) of 45 seconds and the
                    # last delay multiplied by 2.5, rounding to two decimal places.
                    mdelay = min(delay_cap, round((random.uniform(1, mdelay * 2.5)), 2))

                    attempts += 1
                else:
                    raise
        raise RetryLimitExceededError(
            "Exceeded request limit {0} times. Aborting.".format(max_retries)
        )

    return decorated


class ConnectionManager(object):
    """
    The Connection Manager is used to create boto3 clients for
    the various AWS services that Sceptre needs to interact with.

    :param profile: The AWS credentials profile that should be used.
    :param sceptre_role: The sceptre_role that should be assumed in the account.
    :param stack_name: The CloudFormation stack name for this connection.
    :param region: The region to use.
    :param sceptre_role_session_duration: The duration to assume the specified sceptre_role per session.
    """

    # STACK_DEFAULT is a sentinel value meaning "default to the stack's configuration". This is in
    # contrast with passing None, which would mean "use no value".
    STACK_DEFAULT = "[STACK DEFAULT]"

    _session_lock = threading.Lock()
    _client_lock = threading.Lock()
    _boto_sessions = {}
    _clients = {}
    _stack_keys = {}

    iam_role = create_deprecated_alias_property(
        "iam_role", "sceptre_role", "4.0.0", "5.0.0"
    )
    sceptre_role_session_duration = 0
    iam_role_session_duration = create_deprecated_alias_property(
        "iam_role_session_duration", "sceptre_role_session_duration", "4.0.0", "5.0.0"
    )

    def __init__(
        self,
        region: str,
        profile: Optional[str] = None,
        stack_name: Optional[str] = None,
        sceptre_role: Optional[str] = None,
        sceptre_role_session_duration: Optional[int] = None,
        *,
        session_class=boto3.Session,
        get_envs_func=lambda: os.environ,
    ):
        self.logger = logging.getLogger(__name__)

        self.region = region
        self.profile = profile
        self.stack_name = stack_name
        self.sceptre_role = sceptre_role
        self.sceptre_role_session_duration = sceptre_role_session_duration

        if stack_name:
            self._stack_keys[stack_name] = (region, profile, sceptre_role)

        self._session_class = session_class
        self._get_envs = get_envs_func

    def __repr__(self):
        return (
            "sceptre.connection_manager.ConnectionManager(region='{0}', "
            "profile='{1}', stack_name='{2}', sceptre_role='{3}', sceptre_role_session_duration='{4}')".format(
                self.region,
                self.profile,
                self.stack_name,
                self.sceptre_role,
                self.sceptre_role_session_duration,
            )
        )

    def get_session(
        self,
        profile: Optional[str] = STACK_DEFAULT,
        region: Optional[str] = STACK_DEFAULT,
        sceptre_role: Optional[str] = STACK_DEFAULT,
        *,
        iam_role: Optional[str] = STACK_DEFAULT,
    ) -> boto3.Session:
        """
        Returns a boto3 session for the targeted profile, region, and sceptre_role.

        For each of profile, region, and sceptre_role, these values will default to the ConnectionManager's
        configured default values (which correspond to the Stack's configuration). These values can
        be overridden, however, by passing them explicitly.

        :param profile: The name of the AWS Profile as configured in the local environment. Passing
            None will result in no profile being specified. Defaults to the ConnectionManager's
            configured profile (if there is one).
        :param region: The AWS Region the session should be configured with. Defaults to the
            ConnectionManager's configured region.
        :param sceptre_role: The IAM role ARN that is assumed using STS to create the session. Passing
            None will result in no IAM role being assumed. Defaults to the ConnectionManager's
            configured sceptre_role (if there is one).
        :param iam_role: An alias for sceptre_role; Deprecated in v4.0.0 and will be removed in
            v5.0.0.

        :returns: The Boto3 session.
        :raises: botocore.exceptions.ClientError
        """
        profile, region, sceptre_role = self._determine_session_args(
            profile, region, sceptre_role, iam_role
        )

        return self._get_session(profile, region, sceptre_role)

    def _determine_session_args(
        self, profile: str, region: str, sceptre_role: str, iam_role: str
    ) -> Tuple[str, str, str]:
        profile = self.profile if profile == self.STACK_DEFAULT else profile
        region = self.region if region == self.STACK_DEFAULT else region
        sceptre_role = self._coalesce_sceptre_role(iam_role, sceptre_role)
        sceptre_role = (
            self.sceptre_role if sceptre_role == self.STACK_DEFAULT else sceptre_role
        )
        # For historical reasons, if all three values are "None", that means we default to the
        # Stack's configuration.
        if (profile, region, sceptre_role) == (None, None, None):
            profile, region, sceptre_role = self.profile, self.region, self.sceptre_role

        return profile, region, sceptre_role

    def _emit_iam_role_deprecation_warning(self):
        warnings.warn(
            deprecation.DeprecatedWarning(
                "The iam_role parameter", "4.0.0", "5.0.0", "Use sceptre_role instead"
            ),
            DeprecationWarning,
            stacklevel=3,
        )

    def create_session_environment_variables(
        self,
        profile: Optional[str] = STACK_DEFAULT,
        region: Optional[str] = STACK_DEFAULT,
        sceptre_role: Optional[str] = STACK_DEFAULT,
        include_system_envs: bool = True,
    ) -> Dict[str, str]:
        """Creates the standard AWS environment variables that would need to be passed to a
        subprocess in a hook, resolver, or template handler and allow that subprocess to work with
        the currently configured session.

        The environment variables returned by this method should be everything needed for
        subprocesses to properly interact with AWS using the ConnectionManager's configurations for
        profile, sceptre_role, and region. By default, they include the other process environment
        variables, such as PATH and any others. If you do not want the other environment variables,
        you can toggle these off via include_system_envs=False.

        | Notes on including system envs:
        |   * The AWS_DEFAULT_REGION, AWS_REGION, AWS_ACCESS_KEY_ID, and AWS_SECRET_ACCESS_KEY
        |     environment variables (if they are set in the Sceptre process) will be overwritten in
        |     the returned dict with the correct values from the newly created Session.
        |   * If the AWS_SESSION_TOKEN environment variable is currently set for the process, this
        |     will be overwritten with the new session's token (if there is one) or removed from the
        |     returned environment variables dict (if the new session doesn't have a token).

        :param profile: The name of the AWS Profile as configured in the local environment. Passing
            None will result in no profile being specified. Defaults to the ConnectionManager's
            configured profile (if there is one).
        :param region: The AWS Region the session should be configured with. Defaults to the
            ConnectionManager's configured region.
        :param sceptre_role: The IAM role ARN that is assumed using STS to create the session. Passing
            None will result in no IAM role being assumed. Defaults to the ConnectionManager's
            configured sceptre_role (if there is one).
        :param include_system_envs: If True, will return a dict with all the system environment
            variables included. This is useful for creating a complete dict of environment variables
            to pass to a subprocess. If set to False, this method will ONLY return the relevant AWS
            environment variables. Defaults to True.

        :returns: A dict of environment variables with the appropriate credentials available for use.
        """
        session = self.get_session(profile, region, sceptre_role)
        # Set aws environment variables specific to whatever AWS configuration has been set on the
        # stack's connection manager.
        credentials: Credentials = session.get_credentials()
        envs = dict(**self._get_envs()) if include_system_envs else {}

        if include_system_envs:
            # We don't want a profile specified, since that could interfere with the credentials we're
            # about to set. Even if we're using a profile, the credentials will already reflect that
            # profile's configurations.
            envs.pop("AWS_PROFILE", None)

        envs.update(
            AWS_ACCESS_KEY_ID=credentials.access_key,
            AWS_SECRET_ACCESS_KEY=credentials.secret_key,
            # Most AWS SDKs use AWS_DEFAULT_REGION for the region; some use AWS_REGION
            AWS_DEFAULT_REGION=session.region_name,
            AWS_REGION=session.region_name,
        )

        if credentials.token:
            envs["AWS_SESSION_TOKEN"] = credentials.token
        # There might not be a session token, so if there isn't one, make sure it doesn't exist in
        # the envs being passed to the subprocess
        elif include_system_envs:
            envs.pop("AWS_SESSION_TOKEN", None)

        return envs

    def _get_session(
        self,
        profile: Optional[str],
        region: Optional[str],
        sceptre_role: Optional[str],
        *,
        iam_role: Optional[str] = None,
    ) -> boto3.Session:
        if iam_role is not None:
            self._emit_iam_role_deprecation_warning()
            sceptre_role = iam_role

        with self._session_lock:
            self.logger.debug("Getting Boto3 session")
            key = (region, profile, sceptre_role)

            if self._boto_sessions.get(key) is None:
                self.logger.debug("No Boto3 session found, creating one...")
                self.logger.debug("Using cli credentials...")
                environ = self._get_envs()
                # Credentials from env take priority over profile
                config = {
                    "profile_name": profile,
                    "region_name": region,
                    "aws_access_key_id": environ.get("AWS_ACCESS_KEY_ID"),
                    "aws_secret_access_key": environ.get("AWS_SECRET_ACCESS_KEY"),
                    "aws_session_token": environ.get("AWS_SESSION_TOKEN"),
                }

                session = self._session_class(**config)
                self._boto_sessions[key] = session

                if session.get_credentials() is None:
                    raise InvalidAWSCredentialsError(
                        "Session credentials were not found. Profile: {0}. Region: {1}.".format(
                            config["profile_name"], config["region_name"]
                        )
                    )

                if sceptre_role:
                    sts_client = session.client("sts")
                    # maximum session name length is 64 chars. 56 + "-session" = 64
                    session_name = f'{sceptre_role.split("/")[-1][:56]}-session'
                    assume_role_kwargs = {
                        "RoleArn": sceptre_role,
                        "RoleSessionName": session_name,
                    }
                    if self.sceptre_role_session_duration:
                        assume_role_kwargs["DurationSeconds"] = (
                            self.sceptre_role_session_duration
                        )
                    sts_response = sts_client.assume_role(**assume_role_kwargs)

                    credentials = sts_response["Credentials"]
                    session = self._session_class(
                        aws_access_key_id=credentials["AccessKeyId"],
                        aws_secret_access_key=credentials["SecretAccessKey"],
                        aws_session_token=credentials["SessionToken"],
                        region_name=region,
                    )

                    if session.get_credentials() is None:
                        raise InvalidAWSCredentialsError(
                            "Session credentials were not found. Role: {0}. Region: {1}.".format(
                                sceptre_role, region
                            )
                        )

                    self._boto_sessions[key] = session

                self.logger.debug(
                    "Using credential set from %s: %s",
                    session.get_credentials().method,
                    {
                        "AccessKeyId": mask_key(session.get_credentials().access_key),
                        "SecretAccessKey": mask_key(
                            session.get_credentials().secret_key
                        ),
                        "Region": session.region_name,
                    },
                )

                self.logger.debug("Boto3 session created")

            return self._boto_sessions[key]

    def _get_client(self, service, region, profile, stack_name, sceptre_role):
        """
        Returns the Boto3 client associated with <service>.

        Equivalent to calling Boto3.client(<service>). Gets the client using
        ``boto_session``.

        :param service: The Boto3 service to return a client for.
        :type service: str
        :returns: The Boto3 client.
        :rtype: boto3.client.Client
        """
        with self._client_lock:
            key = (service, region, profile, stack_name, sceptre_role)
            if self._clients.get(key) is None:
                self.logger.debug("No %s client found, creating one...", service)
                self._clients[key] = self._get_session(
                    profile, region, sceptre_role
                ).client(service)

            return self._clients[key]

    @_retry_boto_call
    def call(
        self,
        service: str,
        command: str,
        kwargs: Dict[str, Any] = None,
        profile: Optional[str] = STACK_DEFAULT,
        region: Optional[str] = STACK_DEFAULT,
        stack_name: Optional[str] = None,
        sceptre_role: Optional[str] = STACK_DEFAULT,
        *,
        iam_role: Optional[str] = STACK_DEFAULT,
    ):
        """
        Makes a thread-safe Boto3 client call.

        Equivalent to ``boto3.client(<service>).<command>(**kwargs)``.

        | Note regarding the profile, region, and sceptre_role parameters:
        |    We will interpret each parameter individually this way:
        |      * If the value passed is the STACK_DEFAULT constant, we'll assume it to mean we ought
        |        to use the target stack's value of that parameter.
        |      * If the value passed is None, we will interpret that as an explicit request to nullify
        |        the target stack's setting. Note: While this is valid for profile and sceptre_role,
        |        it will likely blow up if doing this for region, since AWS almost always requires that.
        |      * Otherwise, any value that has been specified will override the target stack's
        |        configuration, regardless of what has been passed for other parameters.
        |      * In the case that `None` has been specified for all parameters, that will be
        |        interpreted as using the target stack's values for all three, falling back to the
        |        current stack.

        :param service: The Boto3 service to return a client for.
        :param command: The Boto3 command to call.
        :param kwargs: The keyword arguments to supply to <command>.
        :param profile: The profile to use when invoking the command; Defaults to the stack's configuration
        :param region: The region to use when invoking the command; Default's to the stack's configuration
        :param stack_name: The name of the stack whose configuration to use. Defaults to the current stack
        :param sceptre_role: The IAM Role ARN to assume in order to invoke the command; Defaults to
            the stack's configuration.
        :param iam_role: DEPRECATED. Use sceptre_role instead.
        :returns: The response from the Boto3 call.
        """
        # If stack_name has been specified and we've already cached the region/profile/role
        # configured for that stack, the "defaults" we'll use will be those of that stack rather then
        # the defaults for the current ConnectionManager instance.
        #
        # stack_name is not used often and only really makes sense when we are acting inside Stack A
        # but needing to interact with Stack B using Stack B's configurations. This is mostly only
        # done when we're getting Stack B's outputs to resolve for Stack A's configuration.
        if stack_name and stack_name in self._stack_keys:
            stack_region, stack_profile, stack_sceptre_role = self._stack_keys[
                stack_name
            ]
            sceptre_role = self._coalesce_sceptre_role(iam_role, sceptre_role)
            # For historical/legacy purposes, if `None` is explicitly passed for all three parameters,
            # this will be interpreted to mean we're going to use the profile/region/role configuration
            # of the stack name. This could potentially interfere with an explicit attempt to nullify
            # a setting; However, in that case, we'd need to be setting the region to None... which
            # is unlikely, since that is a required stack configuration. This is the way this
            # function has always operated, so to change this behavior could break or cause
            # unexpected behavior elsewhere.
            if (region, profile, sceptre_role) == (None, None, None):
                region, profile, sceptre_role = (
                    stack_region,
                    stack_profile,
                    stack_sceptre_role,
                )
            # In every other circumstance, we will interpret each parameter individually according
            # to the way described in the docstring.
            else:
                region = stack_region if region == self.STACK_DEFAULT else region
                profile = stack_profile if profile == self.STACK_DEFAULT else profile
                sceptre_role = (
                    stack_sceptre_role
                    if sceptre_role == self.STACK_DEFAULT
                    else sceptre_role
                )
        # In most cases, we won't be targeting another stack's configurations. Instead, we'll want
        # to be using the configurations of the CURRENT stack.
        else:
            profile, region, sceptre_role = self._determine_session_args(
                profile, region, sceptre_role, iam_role
            )

        if kwargs is None:  # pragma: no cover
            kwargs = {}

        client = self._get_client(service, region, profile, stack_name, sceptre_role)
        return getattr(client, command)(**kwargs)

    def _coalesce_sceptre_role(self, iam_role: str, sceptre_role: str) -> str:
        """Evaluates the iam_role and sceptre_role parameters as passed to determine which value to
        use.
        """
        if sceptre_role == self.STACK_DEFAULT and iam_role != self.STACK_DEFAULT:
            self._emit_iam_role_deprecation_warning()
            sceptre_role = iam_role
        return sceptre_role



import os
import errno

import click
import yaml

from sceptre.config.reader import STACK_GROUP_CONFIG_ATTRIBUTES
from sceptre.cli.helpers import catch_exceptions
from sceptre.exceptions import ProjectAlreadyExistsError


@click.group(name="new")
def new_group():
    """
    Commands for initialising Sceptre projects.

    """
    pass


@new_group.command(
    "group", short_help="Creates a new Stack Group directory in a project."
)
@click.argument("stack_group")
@catch_exceptions
@click.pass_context
def new_stack_group(ctx, stack_group):
    """
    Creates StackGroup folder in the project and a config.yaml with any
    required properties.
    \f

    :param stack_group: Name of the StackGroup directory to create.
    :type stack_group: str
    """
    cwd = ctx.obj.get("project_path")
    for item in os.listdir(cwd):
        # If already a config folder create a sub stack_group
        if os.path.isdir(item) and item == "config":
            config_dir = os.path.join(os.getcwd(), "config")
            _create_new_stack_group(config_dir, stack_group)


@new_group.command("project", short_help="Creates a new project.")
@catch_exceptions
@click.argument("project_name")
@click.pass_context
def new_project(ctx, project_name):
    """
    Creates PROJECT_NAME project folder and a config.yaml with any
    required properties.
    \f

    :param project_name: The name of the Sceptre Project to create.
    :type project_name: str
    """
    cwd = os.getcwd()
    sceptre_folders = {"config", "templates"}
    project_folder = os.path.join(cwd, project_name)
    try:
        os.mkdir(project_folder)
    except OSError as e:
        # Check if stack_group folder already exists
        if e.errno == errno.EEXIST:
            raise ProjectAlreadyExistsError(
                'Folder "{0}" already exists.'.format(project_name)
            )
        else:
            raise

    for folder in sceptre_folders:
        folder_path = os.path.join(project_folder, folder)
        os.makedirs(folder_path)

    defaults = {
        "project_code": project_name,
        "region": os.environ.get("AWS_DEFAULT_REGION", ""),
    }

    config_path = os.path.join(cwd, project_name, "config")
    _create_config_file(config_path, config_path, defaults)


def _create_new_stack_group(config_dir, new_path):
    """
    Creates the subfolder for the stack_group specified by `new_path`
    starting from the `config_dir`. Even if folder path already exists,
    they want to initialise `config.yaml`.

    :param config_dir: The directory path to the top-level config folder.
    :type config_dir: str
    :param new_path: The directory path to the stack_group folder.
    :type new_path: str
    """
    # Create full path to stack_group
    folder_path = os.path.join(config_dir, new_path)
    new_config_msg = "Do you want initialise config.yaml?"

    # Make folders for the stack_group
    try:
        os.makedirs(folder_path)
    except OSError as e:
        # Check if stack_group folder already exists
        if e.errno == errno.EEXIST:
            new_config_msg = "StackGroup path exists. " + new_config_msg
        else:
            raise

    if click.confirm(new_config_msg):
        _create_config_file(config_dir, folder_path)


def _get_nested_config(config_dir, path):
    """
    Collects nested config from between `config_dir` and `path`. Config at
    lower level as greater precedence.

    :param config_dir: The directory path to the top-level config folder.
    :type config_dir: str
    :param path: The directory path to the stack_group folder.
    :type path: str
    :returns: The nested config.
    :rtype: dict
    """
    config = {}
    for root, _, files in os.walk(config_dir, followlinks=True):
        # Check that folder is within the final stack_group path
        if path.startswith(root) and "config.yaml" in files:
            config_path = os.path.join(root, "config.yaml")
            with open(config_path) as config_file:
                config.update(yaml.safe_load(config_file))
    return config


def _create_config_file(config_dir, path, defaults={}):
    """
    Creates a `config.yaml` file in the given path. The user is asked for
    values for requried properties. Defaults are suggested with values in
    `defaults` and then vaules found in parent `config.yaml` files. If
    properties and their values are the same as in parent `config.yaml`, then
    they are not included. No file is produced if require values are satisfied
    by parent `config.yaml` files.

    :param config_dir: The directory path to the top-level config folder.
    :type config_dir: str
    :param path: The directory path to the stack_group folder.
    :type path: str
    :param defaults: Defaults to present to the user for config.
    :type defaults: dict
    """
    config = dict.fromkeys(STACK_GROUP_CONFIG_ATTRIBUTES.required, "")
    parent_config = _get_nested_config(config_dir, path)

    # Add standard defaults
    config.update(defaults)

    # Add parent config values as defaults
    config.update(parent_config)

    # Ask for new values
    for key, value in config.items():
        config[key] = click.prompt("Please enter a {0}".format(key), default=value)

    # Remove values if parent config are the same
    config = {k: v for k, v in config.items() if parent_config.get(k) != v}

    # Write config.yaml if config not empty
    filepath = os.path.join(path, "config.yaml")
    if config:
        with open(filepath, "w") as config_file:
            yaml.safe_dump(config, stream=config_file, default_flow_style=False)
    else:
        click.echo("No config.yaml file needed - covered by parent config.")



import click

from sceptre.context import SceptreContext
from sceptre.cli.helpers import catch_exceptions, write
from sceptre.plan.plan import SceptrePlan


@click.command(name="status", short_help="Print status of stack or stack_group.")
@click.argument("path")
@click.pass_context
@catch_exceptions
def status_command(ctx, path):
    """
    Prints the stack status or the status of the stacks within a
    stack_group for a given config PATH.
    \f

    :param path: Path to execute the command on.
    :type path: str
    """
    context = SceptreContext(
        command_path=path,
        command_params=ctx.params,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        options=ctx.obj.get("options"),
        no_colour=ctx.obj.get("no_colour"),
        output_format=ctx.obj.get("output_format"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies"),
    )

    plan = SceptrePlan(context)
    responses = plan.get_status()
    message = "\n".join(
        "{}: {}".format(stack.name, status) for stack, status in responses.items()
    )
    write(message, no_colour=context.no_colour)



from sceptre.resolvers import Resolver


class Sub(Resolver):
    """This resolver allows you to create a string using Python string format syntax. This is a
    great way to combine together a number of resolver outputs into a single string. This functions
    very similarly to Cloudformation's ``!Sub`` intrinsic function.

    The argument to this resolver should be a two-element list: (1) Is the format string, using
    curly-brace templates to indicate variables, and (2) a dictionary where the keys are the format
    string's variable names and the values are the variable values.

    Example:

       parameters:
         ConnectionString: !sub
           - "postgres://{username}:{password}@{hostname}:{port}/{database}"
           - username: {{ var.username }}
             password: !ssm /my/ssm/password
             hostname: !stack_output my/database/stack.yaml::HostName
             port: !stack_output my/database/stack.yaml::Port
             database: {{var.database}}
    """

    def resolve(self):
        error_message = (
            "The argument to !sub must be a two-element list, where the first element is the "
            "a format string and the second element is a dict of values to interpolate into it."
        )
        if not isinstance(self.argument, list) or len(self.argument) != 2:
            self.raise_invalid_argument_error(error_message)

        template, variables = self.argument
        if not isinstance(template, str) or not isinstance(variables, dict):
            self.raise_invalid_argument_error(error_message)

        try:
            return template.format(**variables)
        except KeyError as e:
            self.raise_invalid_argument_error(
                f"Could not find !sub argument for {e}", e
            )



from uuid import uuid1

import click

from typing import Optional
from sceptre.context import SceptreContext
from sceptre.cli.helpers import catch_exceptions, confirmation
from sceptre.cli.helpers import write, stack_status_exit_code
from sceptre.cli.helpers import simplify_change_set_description
from sceptre.stack_status import StackChangeSetStatus
from sceptre.plan.plan import SceptrePlan


@click.command(name="update", short_help="Update a stack.")
@click.argument("path")
@click.option(
    "-c", "--change-set", is_flag=True, help="Create a change set before updating."
)
@click.option("-v", "--verbose", is_flag=True, help="Display verbose output.")
@click.option("-y", "--yes", is_flag=True, help="Assume yes to all questions.")
@click.option(
    "--disable-rollback/--enable-rollback",
    default=None,
    help="Disable or enable the cloudformation automatic rollback",
)
@click.pass_context
@catch_exceptions
def update_command(
    ctx, path, change_set, verbose, yes, disable_rollback: Optional[bool]
):
    """
    Updates a stack for a given config PATH. Or perform an update via
    change-set when the change-set flag is set.
    \f

    :param path: Path to execute the command on.
    :type path: str
    :param change_set: Whether a change set should be created.
    :type change_set: bool
    :param verbose: A flag to print a verbose output.
    :type verbose: bool
    :param yes: A flag to answer 'yes' to all CLI questions.
    :type yes: bool
    :param disable_rollback: A flag to disable cloudformation rollback.
    """

    context = SceptreContext(
        command_path=path,
        command_params=ctx.params,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        options=ctx.obj.get("options"),
        output_format=ctx.obj.get("output_format"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies"),
    )

    plan = SceptrePlan(context)

    if change_set:
        change_set_name = "-".join(["change-set", uuid1().hex])
        plan.create_change_set(change_set_name)
        try:
            # Wait for change set to be created
            statuses = plan.wait_for_cs_completion(change_set_name)

            at_least_one_ready = False

            for status in statuses.values():
                # Exit if change set fails to create
                if status not in (
                    StackChangeSetStatus.READY,
                    StackChangeSetStatus.NO_CHANGES,
                ):
                    write("Failed to create change set", context.output_format)
                    exit(1)

                if status == StackChangeSetStatus.READY:
                    at_least_one_ready = True

            # If none are ready, and we haven't exited, there are no changes
            if not at_least_one_ready:
                write("No changes detected", context.output_format)
                exit(0)

            # Describe changes
            descriptions = plan.describe_change_set(change_set_name)
            for stack, description in descriptions.items():
                # No need to print if there are no changes
                if statuses[stack] == StackChangeSetStatus.NO_CHANGES:
                    continue

                if not verbose:
                    description = simplify_change_set_description(description)
                write(description, context.output_format)

            # Execute change set if happy with changes
            if yes or click.confirm("Proceed with stack update?"):
                plan.execute_change_set(change_set_name)

        finally:
            # Clean up by deleting change set
            plan.delete_change_set(change_set_name)
    else:
        confirmation("update", yes, command_path=path)
        responses = plan.update()
        exit(stack_status_exit_code(responses.values()))



from logging import LoggerAdapter, Logger
from typing import MutableMapping, Any, Tuple


class StackLoggerAdapter(LoggerAdapter):
    def __init__(self, logger: Logger, stack_name: str, extra: dict = None):
        """A small wrapper around a Logger that prefixes log messages with the stack name.

        :param logger: The logger to wrap
        :param stack_name: The name of the stack to every log message
        :param extra: Extra kwargs to add to the log context (if any)
        """
        super().__init__(logger, extra or {})
        self.stack_name = stack_name

    def process(
        self, msg: str, kwargs: MutableMapping[str, Any]
    ) -> Tuple[Any, MutableMapping[str, Any]]:
        msg = f"{self.stack_name} - {msg}"
        return super().process(msg, kwargs)



# -*- coding: utf-8 -*-

"""
sceptre.stack

This module implements a Stack class, which stores a Stack's data.

"""

import logging

from typing import List, Dict, Union, Any, Optional
from deprecation import deprecated

from sceptre import __version__
from sceptre.connection_manager import ConnectionManager
from sceptre.exceptions import InvalidConfigFileError
from sceptre.helpers import (
    get_external_stack_name,
    sceptreise_path,
    create_deprecated_alias_property,
)
from sceptre.hooks import Hook, HookProperty
from sceptre.resolvers import (
    ResolvableContainerProperty,
    ResolvableValueProperty,
    RecursiveResolve,
    PlaceholderType,
    Resolver,
)
from sceptre.template import Template


class Stack:
    """
    Stack stores information about a particular CloudFormation Stack.

    :param name: The name of the Stack.

    :param project_code: A code which is prepended to the Stack names\
            of all Stacks built by Sceptre.

    :param template_path: The relative path to the CloudFormation, Jinja2,
            or Python template to build the Stack from. If this is filled,
            `template_handler_config` should not be filled. This field has been deprecated since
            version 4.0.0 and will be removed eventually.

    :param template_handler_config: Configuration for a Template Handler that can resolve
            its arguments to a template string. Should contain the `type` property to specify
            the type of template handler to load. Conflicts with `template_path`.

    :param region: The AWS region to build Stacks in.

    :param template_bucket_name: The name of the S3 bucket the Template is uploaded to.

    :param template_key_prefix: A prefix to the key used to store templates uploaded to S3

    :param required_version: A PEP 440 compatible version specifier. If the Sceptre version does\
            not fall within the given version requirement it will abort.

    :param parameters: The keys must match up with the name of the parameter.\
            The value must be of the type as defined in the template.

    :param sceptre_user_data: Data passed into\
            `sceptre_handler(sceptre_user_data)` function in Python templates\
            or accessible under `sceptre_user_data` variable within Jinja2\
            templates.

    :param hooks: A list of arbitrary shell or python commands or scripts to\
            run.

    :param s3_details: Details used for uploading templates to S3.

    :param dependencies: The relative path to the Stack, including the file\
            extension of the Stack.

    :param cloudformation_service_role: The ARN of a CloudFormation Service Role that is assumed\
            by CloudFormation to create, update or delete resources.

    :param protected: Stack protection against execution.

    :param tags: CloudFormation Tags to be applied to the Stack.

    :param external_name: The real stack name used for CloudFormation

    :param notifications: SNS topic ARNs to publish Stack related events to.\
            A maximum of 5 ARNs can be specified per Stack.

    :param on_failure: This parameter describes the action taken by\
            CloudFormation when a Stack fails to create.

    :param disable_rollback: If True, cloudformation will not rollback on deployment failures

    :param iam_role: The ARN of a role for Sceptre to assume before interacting
            with the environment. If not supplied, Sceptre uses the user's AWS CLI
            credentials. This field has been deprecated since version 4.0.0 and will be removed
            eventually.

    :param sceptre_role: The ARN of a role for Sceptre to assume before interacting\
            with the environment. If not supplied, Sceptre uses the user's AWS CLI\
            credentials.

    :param iam_role_session_duration: The duration in seconds of the assumed IAM role session.
            This field has been deprecated since version 4.0.0 and will be removed eventually.

    :param sceptre_role_session_duration: The duration in seconds of the assumed IAM role session.

    :param profile: The name of the profile as defined in ~/.aws/config and\
            ~/.aws/credentials.

    :param stack_timeout: A timeout in minutes before considering the Stack\
            deployment as failed. After the specified timeout, the Stack will\
            be rolled back. Specifying zero, as well as omitting the field,\
            will result in no timeout. Supports only positive integer value.

    :param ignore: If True, this stack will be ignored during launches (but it can be explicitly
            deployed with create, update, and delete commands.

    :param obsolete: If True, this stack will operate the same as if ignore was set, but it will
            also be deleted if the prune command is invoked or the --prune option is used with the
            launch command.

    :param sceptre_role_session_duration: The session duration when Scetre assumes a role.\
           If not supplied, Sceptre uses default value (3600 seconds)

    :param stack_group_config: The StackGroup config for the Stack

    :param config: The complete config for the stack. Used by dump config.
    """

    parameters = ResolvableContainerProperty("parameters")
    sceptre_user_data = ResolvableContainerProperty(
        "sceptre_user_data", PlaceholderType.alphanum
    )
    notifications = ResolvableContainerProperty("notifications")
    tags = ResolvableContainerProperty("tags")
    # placeholder_override=None here means that if the template_bucket_name is a resolver,
    # placeholders have been enabled, and that stack hasn't been deployed yet, commands that would
    # otherwise attempt to upload the template (like validate) won't actually use the template bucket
    # and will act as if there was no template bucket set.
    s3_details = ResolvableContainerProperty("s3_details", PlaceholderType.none)
    template_handler_config = ResolvableContainerProperty(
        "template_handler_config", PlaceholderType.alphanum
    )

    template_bucket_name = ResolvableValueProperty(
        "template_bucket_name", PlaceholderType.none
    )
    # Similarly, the placeholder_override=None for sceptre_role means that actions that would otherwise
    # use the sceptre_role will act as if there was no iam role when the sceptre_role stack has not been
    # deployed for commands that allow placeholders (like validate).
    sceptre_role = ResolvableValueProperty("sceptre_role", PlaceholderType.none)
    cloudformation_service_role = ResolvableValueProperty("cloudformation_service_role")

    hooks = HookProperty("hooks")

    iam_role = create_deprecated_alias_property(
        "iam_role",
        "sceptre_role",
        deprecated_in="4.0.0",
        removed_in=None,
    )
    role_arn = create_deprecated_alias_property(
        "role_arn",
        "cloudformation_service_role",
        deprecated_in="4.0.0",
        removed_in=None,
    )
    sceptre_role_session_duration = None
    iam_role_session_duration = create_deprecated_alias_property(
        "iam_role_session_duration",
        "sceptre_role_session_duration",
        deprecated_in="4.0.0",
        removed_in=None,
    )

    def __init__(
        self,
        name: str,
        project_code: str,
        region: str,
        template_path: str = None,
        template_handler_config: dict = None,
        template_bucket_name: str = None,
        template_key_prefix: str = None,
        required_version: str = None,
        parameters: dict = None,
        sceptre_user_data: dict = None,
        hooks: Hook = None,
        s3_details: dict = None,
        sceptre_role: str = None,
        iam_role: str = None,
        dependencies: List["Stack"] = None,
        cloudformation_service_role: str = None,
        role_arn: str = None,
        protected: bool = False,
        tags: dict = None,
        external_name: str = None,
        notifications: List[str] = None,
        on_failure: str = None,
        disable_rollback=False,
        profile: str = None,
        stack_timeout: int = 0,
        sceptre_role_session_duration: Optional[int] = None,
        iam_role_session_duration: Optional[int] = None,
        ignore=False,
        obsolete=False,
        stack_group_config: dict = None,
        config: dict = None,
    ):
        self.logger = logging.getLogger(__name__)

        self.name = sceptreise_path(name)
        self.project_code = project_code
        self.region = region
        self.required_version = required_version
        self.external_name = external_name or get_external_stack_name(
            self.project_code, self.name
        )
        self.dependencies = dependencies or []
        self.protected = protected
        self.on_failure = on_failure
        self.disable_rollback = self._ensure_boolean(
            "disable_rollback", disable_rollback
        )
        self.stack_group_config = stack_group_config or {}
        self.config = config or {}
        self.stack_timeout = stack_timeout
        self.profile = profile
        self.template_key_prefix = template_key_prefix
        self._set_field_with_deprecated_alias(
            "sceptre_role_session_duration",
            sceptre_role_session_duration,
            "iam_role_session_duration",
            iam_role_session_duration,
        )
        self.ignore = self._ensure_boolean("ignore", ignore)
        self.obsolete = self._ensure_boolean("obsolete", obsolete)

        self._template = None
        self._connection_manager = None

        # Resolvers and hooks need to be assigned last
        self.s3_details = s3_details
        self._set_field_with_deprecated_alias(
            "sceptre_role", sceptre_role, "iam_role", iam_role
        )
        self.tags = tags or {}
        self._set_field_with_deprecated_alias(
            "cloudformation_service_role",
            cloudformation_service_role,
            "role_arn",
            role_arn,
        )
        self.template_bucket_name = template_bucket_name
        self._set_field_with_deprecated_alias(
            "template_handler_config",
            template_handler_config,
            "template_path",
            template_path,
            required=True,
            preferred_config_name="template",
        )

        self.s3_details = s3_details
        self.parameters = self._cast_parameters(parameters or {})
        self.sceptre_user_data = sceptre_user_data or {}
        self.notifications = notifications or []

        self.hooks = hooks or {}

    def _ensure_boolean(self, config_name: str, value: Any) -> bool:
        if not isinstance(value, bool):
            raise InvalidConfigFileError(
                f"{self.name}: Value for {config_name} must be a boolean, not a {type(value).__name__}"
            )
        return value

    def _cast_parameters(
        self, parameters: Dict[str, Any]
    ) -> Dict[str, Union[str, List[Union[str, Resolver]], Resolver]]:
        """Cast CloudFormation parameters to valid types"""

        def cast_value(value: Any) -> Union[str, List[Union[str, Resolver]], Resolver]:
            if isinstance(value, bool):
                return "true" if value else "false"
            elif isinstance(value, (int, float)):
                return str(value)
            elif isinstance(value, list):
                return [cast_value(item) for item in value]
            elif isinstance(value, Resolver):
                return value
            return value

        def is_valid(value: Any) -> bool:
            return (
                isinstance(value, str)
                or (
                    isinstance(value, list)
                    and all(
                        isinstance(item, str) or isinstance(item, Resolver)
                        for item in value
                    )
                )
                or isinstance(value, Resolver)
            )

        if not isinstance(parameters, dict):
            raise InvalidConfigFileError(
                f"{self.name}: parameters must be a dictionary of key-value pairs, got {parameters}"
            )

        casted_parameters = {k: cast_value(v) for k, v in parameters.items()}

        if not all(is_valid(value) for value in casted_parameters.values()):
            raise InvalidConfigFileError(
                f"{self.name}: Values for parameters must be strings, lists or resolvers, got {casted_parameters}"
            )

        return casted_parameters

    def __repr__(self):
        return (
            "sceptre.stack.Stack("
            f"name='{self.name}', "
            f"project_code={self.project_code}, "
            f"template_handler_config={self.template_handler_config}, "
            f"region={self.region}, "
            f"template_bucket_name={self.template_bucket_name}, "
            f"template_key_prefix={self.template_key_prefix}, "
            f"required_version={self.required_version}, "
            f"sceptre_role={self.sceptre_role}, "
            f"sceptre_role_session_duration={self.sceptre_role_session_duration}, "
            f"profile={self.profile}, "
            f"sceptre_user_data={self.sceptre_user_data}, "
            f"parameters={self.parameters}, "
            f"hooks={self.hooks}, "
            f"s3_details={self.s3_details}, "
            f"dependencies={self.dependencies}, "
            f"cloudformation_service_role={self.cloudformation_service_role}, "
            f"protected={self.protected}, "
            f"tags={self.tags}, "
            f"external_name={self.external_name}, "
            f"notifications={self.notifications}, "
            f"on_failure={self.on_failure}, "
            f"disable_rollback={self.disable_rollback}, "
            f"stack_timeout={self.stack_timeout}, "
            f"stack_group_config={self.stack_group_config}, "
            f"ignore={self.ignore}, "
            f"obsolete={self.obsolete}"
            ")"
        )

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(str(self))

    @property
    def connection_manager(self) -> ConnectionManager:
        """Returns the ConnectionManager for the stack, creating it if it has not yet been created.

        :returns: ConnectionManager.
        """
        if self._connection_manager is None:
            cache_connection_manager = True
            try:
                sceptre_role = self.sceptre_role
            except RecursiveResolve:
                # This would be the case when sceptre_role is set with a resolver (especially stack_output)
                # that uses the stack's connection manager. This creates a temporary condition where
                # you need the iam role to get the iam role. To get around this, it will temporarily
                # use None as the sceptre_role but will re-attempt to resolve the value in future accesses.
                # Since the Stack Output resolver (the most likely culprit) uses the target stack's
                # sceptre_role rather than the current stack's one anyway, it actually doesn't matter,
                # since the stack defining that sceptre_role won't actually be using that sceptre_role.
                self.logger.debug(
                    "Resolving sceptre_role requires the Stack connection manager. Temporarily setting "
                    "the sceptre_role to None until it can be fully resolved."
                )
                sceptre_role = None
                cache_connection_manager = False

            connection_manager = ConnectionManager(
                self.region,
                self.profile,
                self.external_name,
                sceptre_role,
                self.sceptre_role_session_duration,
            )
            if cache_connection_manager:
                self._connection_manager = connection_manager
            else:  # Return early without caching the connection manager.
                return connection_manager

        return self._connection_manager

    @property
    def template(self):
        """
        Returns the CloudFormation Template used to create the Stack.

        :returns: The Stack's template.
        :rtype: Template
        """
        if self._template is None:
            self._template = Template(
                name=self.name,
                handler_config=self.template_handler_config,
                sceptre_user_data=self.sceptre_user_data,
                stack_group_config=self.stack_group_config,
                s3_details=self.s3_details,
                connection_manager=self.connection_manager,
            )
        return self._template

    @property
    @deprecated(
        deprecated_in="4.0.0",
        removed_in=None,
        current_version=__version__,
        details="Use the template Stack Config key instead.",
    )
    def template_path(self) -> str:
        """The path argument from the template_handler config. This field is deprecated as of v4.0.0
        and will be removed in v5.0.0.
        """
        return self.template_handler_config["path"]

    @template_path.setter
    @deprecated(
        deprecated_in="4.0.0",
        removed_in=None,
        current_version=__version__,
        details="Use the template Stack Config key instead.",
    )
    def template_path(self, value: str):
        self.template_handler_config = {"type": "file", "path": value}

    def _set_field_with_deprecated_alias(
        self,
        preferred_attribute_name,
        preferred_value,
        deprecated_attribute_name,
        deprecated_value,
        *,
        required=False,
        preferred_config_name=None,
        deprecated_config_name=None,
    ):
        # This is a generic truthiness check. All current default values are falsy, so this should work.
        # If we ever use this function where the default value is NOT falsy, this will be a problem.
        preferred_config_name = preferred_config_name or preferred_attribute_name
        deprecated_config_name = deprecated_config_name or deprecated_attribute_name

        if preferred_value and deprecated_value:
            raise InvalidConfigFileError(
                f"Both '{preferred_config_name}' and '{deprecated_config_name}' are set. You should only set a "
                f"value for {preferred_config_name} because {deprecated_config_name} is deprecated."
            )
        elif preferred_value:
            setattr(self, preferred_attribute_name, preferred_value)
        elif deprecated_value:
            setattr(self, deprecated_attribute_name, deprecated_value)
        elif required:
            raise InvalidConfigFileError(
                f"{preferred_config_name} is a required Stack Config."
            )
        else:  # In case they're both falsy, we should just set the value using the preferred value.
            setattr(self, preferred_attribute_name, preferred_value)



# -*- coding: utf-8 -*-

"""
sceptre.plan.actions

This module implements the StackActions class which provides the functionality
available to a Stack.
"""

import json
import logging
import time
import typing
import urllib
import botocore

from datetime import datetime, timedelta
from dateutil.tz import tzutc
from os import path

from sceptre.connection_manager import ConnectionManager

from sceptre.exceptions import (
    CannotUpdateFailedStackError,
    ProtectedStackError,
    StackDoesNotExistError,
    UnknownStackChangeSetStatusError,
    UnknownStackStatusError,
)
from sceptre.helpers import extract_datetime_from_aws_response_headers
from sceptre.hooks import add_stack_hooks, add_stack_hooks_with_aliases
from sceptre.stack import Stack
from sceptre.stack_status import StackChangeSetStatus, StackStatus

from typing import Dict, Optional, Tuple, Union

if typing.TYPE_CHECKING:
    from sceptre.diffing.stack_differ import StackDiff, StackDiffer


class StackActions:
    """
    StackActions stores the operations a Stack can take, such as creating or
    deleting the Stack.

    :param stack: A Stack object
    :type stack: sceptre.stack.Stack
    """

    def __init__(self, stack: Stack):
        self.stack = stack
        self.name = self.stack.name
        self.logger = logging.getLogger(__name__)
        self.connection_manager = ConnectionManager(
            self.stack.region,
            self.stack.profile,
            self.stack.external_name,
            self.stack.sceptre_role,
            self.stack.sceptre_role_session_duration,
        )

    @add_stack_hooks
    def create(self):
        """
        Creates a Stack.

        :returns: The Stack's status.
        :rtype: sceptre.stack_status.StackStatus
        """
        self._protect_execution()
        self.logger.info("%s - Creating Stack", self.stack.name)
        create_stack_kwargs = {
            "StackName": self.stack.external_name,
            "Parameters": self._format_parameters(self.stack.parameters),
            "Capabilities": [
                "CAPABILITY_IAM",
                "CAPABILITY_NAMED_IAM",
                "CAPABILITY_AUTO_EXPAND",
            ],
            "NotificationARNs": self.stack.notifications,
            "Tags": [
                {"Key": str(k), "Value": str(v)} for k, v in self.stack.tags.items()
            ],
        }

        # can specify either DisableRollback or OnFailure , but not both
        if self.stack.disable_rollback:
            create_stack_kwargs.update({"DisableRollback": self.stack.disable_rollback})
        elif self.stack.on_failure:
            create_stack_kwargs.update({"OnFailure": self.stack.on_failure})

        create_stack_kwargs.update(self.stack.template.get_boto_call_parameter())
        create_stack_kwargs.update(self._get_role_arn())
        create_stack_kwargs.update(self._get_stack_timeout())

        try:
            response = self.connection_manager.call(
                service="cloudformation",
                command="create_stack",
                kwargs=create_stack_kwargs,
            )

            self.logger.debug(
                "%s - Create stack response: %s", self.stack.name, response
            )

            status = self._wait_for_completion(boto_response=response)
        except botocore.exceptions.ClientError as exp:
            if exp.response["Error"]["Code"] == "AlreadyExistsException":
                self.logger.info("%s - Stack already exists", self.stack.name)

                status = StackStatus.COMPLETE
            else:
                raise

        return status

    @add_stack_hooks
    def update(self):
        """
        Updates the Stack.

        :returns: The Stack's status.
        :rtype: sceptre.stack_status.StackStatus
        """
        self._protect_execution()
        self.logger.info("%s - Updating Stack", self.stack.name)
        try:
            update_stack_kwargs = {
                "StackName": self.stack.external_name,
                "Parameters": self._format_parameters(self.stack.parameters),
                "Capabilities": [
                    "CAPABILITY_IAM",
                    "CAPABILITY_NAMED_IAM",
                    "CAPABILITY_AUTO_EXPAND",
                ],
                "NotificationARNs": self.stack.notifications,
                "Tags": [
                    {"Key": str(k), "Value": str(v)} for k, v in self.stack.tags.items()
                ],
            }

            if self.stack.disable_rollback:
                update_stack_kwargs.update(
                    {"DisableRollback": self.stack.disable_rollback}
                )

            update_stack_kwargs.update(self.stack.template.get_boto_call_parameter())
            update_stack_kwargs.update(self._get_role_arn())
            response = self.connection_manager.call(
                service="cloudformation",
                command="update_stack",
                kwargs=update_stack_kwargs,
            )
            status = self._wait_for_completion(
                self.stack.stack_timeout, boto_response=response
            )
            self.logger.debug(
                "%s - Update Stack response: %s", self.stack.name, response
            )

            # Cancel update after timeout
            if status == StackStatus.IN_PROGRESS:
                status = self.cancel_stack_update()

            return status
        except botocore.exceptions.ClientError as exp:
            error_message = exp.response["Error"]["Message"]
            if error_message == "No updates are to be performed.":
                self.logger.info("%s - No updates to perform.", self.stack.name)
                return StackStatus.COMPLETE
            else:
                raise

    def cancel_stack_update(self):
        """
        Cancels a Stack update.

        :returns: The cancelled Stack status.
        :rtype: sceptre.stack_status.StackStatus
        """
        self.logger.warning(
            "%s - Update Stack time exceeded the specified timeout", self.stack.name
        )
        response = self.connection_manager.call(
            service="cloudformation",
            command="cancel_update_stack",
            kwargs={"StackName": self.stack.external_name},
        )
        self.logger.debug(
            "%s - Cancel update Stack response: %s", self.stack.name, response
        )
        return self._wait_for_completion(boto_response=response)

    @add_stack_hooks
    def launch(self) -> StackStatus:
        """
        Launches the Stack.

        If the Stack status is create_failed or rollback_complete, the
        Stack is deleted. Launch then tries to create or update the Stack,
        depending if it already exists. If there are no updates to be
        performed, launch exits gracefully.

        :returns: The Stack's status.
        """
        self._protect_execution()
        self.logger.info(f"{self.stack.name} - Launching Stack")

        try:
            existing_status = self._get_status()
        except StackDoesNotExistError:
            existing_status = "PENDING"

        self.logger.info(
            "%s - Stack is in the %s state", self.stack.name, existing_status
        )

        if existing_status == "PENDING":
            status = self.create()
        elif existing_status in [
            "CREATE_FAILED",
            "ROLLBACK_COMPLETE",
            "REVIEW_IN_PROGRESS",
        ]:
            self.delete()
            status = self.create()
        elif existing_status.endswith("COMPLETE"):
            status = self.update()
        elif existing_status.endswith("IN_PROGRESS"):
            self.logger.info(
                "%s - Stack action is already in progress state and cannot "
                "be updated",
                self.stack.name,
            )
            status = StackStatus.IN_PROGRESS
        elif existing_status.endswith("FAILED"):
            raise CannotUpdateFailedStackError(
                "'{0}' is in a the state '{1}' and cannot be updated".format(
                    self.stack.name, existing_status
                )
            )
        else:
            raise UnknownStackStatusError("{0} is unknown".format(existing_status))
        return status

    @add_stack_hooks
    def delete(self):
        """
        Deletes the Stack.

        :returns: The Stack's status.
        :rtype: sceptre.stack_status.StackStatus
        """
        self._protect_execution()

        self.logger.info("%s - Deleting stack", self.stack.name)
        try:
            status = self._get_status()
        except StackDoesNotExistError:
            self.logger.info("%s - Does not exist.", self.stack.name)
            status = StackStatus.COMPLETE
            return status

        delete_stack_kwargs = {"StackName": self.stack.external_name}
        delete_stack_kwargs.update(self._get_role_arn())
        response = self.connection_manager.call(
            service="cloudformation", command="delete_stack", kwargs=delete_stack_kwargs
        )

        try:
            status = self._wait_for_completion(boto_response=response)
        except StackDoesNotExistError:
            status = StackStatus.COMPLETE
        except botocore.exceptions.ClientError as error:
            if error.response["Error"]["Message"].endswith("does not exist"):
                status = StackStatus.COMPLETE
            else:
                raise
        self.logger.info("%s - delete %s", self.stack.name, status)
        return status

    def lock(self):
        """
        Locks the Stack by applying a deny-all updates Stack Policy.
        """
        policy_path = path.join(
            # need to get to the base install path. __file__ will take us into
            # sceptre/actions so need to walk up the path.
            path.abspath(path.join(__file__, "..", "..")),
            "stack_policies/lock.json",
        )
        self.set_policy(policy_path)
        self.logger.info("%s - Successfully locked Stack", self.stack.name)

    def unlock(self):
        """
        Unlocks the Stack by applying an allow-all updates Stack Policy.
        """
        policy_path = path.join(
            # need to get to the base install path. __file__ will take us into
            # sceptre/actions so need to walk up the path.
            path.abspath(path.join(__file__, "..", "..")),
            "stack_policies/unlock.json",
        )
        self.set_policy(policy_path)
        self.logger.info("%s - Successfully unlocked Stack", self.stack.name)

    def describe(self):
        """
        Returns the a description of the Stack.

        :returns: A Stack description.
        :rtype: dict
        """
        return self.connection_manager.call(
            service="cloudformation",
            command="describe_stacks",
            kwargs={"StackName": self.stack.external_name},
        )

    def describe_events(self):
        """
        Returns the CloudFormation events for a Stack.

        :returns: CloudFormation events for a Stack.
        :rtype: dict
        """
        return self.connection_manager.call(
            service="cloudformation",
            command="describe_stack_events",
            kwargs={"StackName": self.stack.external_name},
        )

    def describe_resources(self):
        """
        Returns the logical and physical resource IDs of the Stack's resources.

        :returns: Information about the Stack's resources.
        :rtype: dict
        """
        self.logger.debug("%s - Describing stack resources", self.stack.name)
        try:
            response = self.connection_manager.call(
                service="cloudformation",
                command="describe_stack_resources",
                kwargs={"StackName": self.stack.external_name},
            )
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Message"].endswith("does not exist"):
                return {self.stack.name: []}
            raise

        self.logger.debug(
            "%s - Describe Stack resource response: %s", self.stack.name, response
        )

        desired_properties = ["LogicalResourceId", "PhysicalResourceId"]

        formatted_response = {
            self.stack.name: [
                {k: v for k, v in item.items() if k in desired_properties}
                for item in response["StackResources"]
            ]
        }
        return formatted_response

    def describe_outputs(self):
        """
        Returns the Stack's outputs.

        :returns: The stack's outputs.
        :rtype: list
        """
        self.logger.debug("%s - Describing stack outputs", self.stack.name)
        response = self.describe()

        return {self.stack.name: response["Stacks"][0].get("Outputs", [])}

    def continue_update_rollback(self):
        """
        Rolls back a Stack in the UPDATE_ROLLBACK_FAILED state to
        UPDATE_ROLLBACK_COMPLETE.
        """
        self.logger.debug("%s - Continuing update rollback", self.stack.name)
        continue_update_rollback_kwargs = {"StackName": self.stack.external_name}
        continue_update_rollback_kwargs.update(self._get_role_arn())
        self.connection_manager.call(
            service="cloudformation",
            command="continue_update_rollback",
            kwargs=continue_update_rollback_kwargs,
        )
        self.logger.info(
            "%s - Successfully initiated continuation of update rollback",
            self.stack.name,
        )

    def set_policy(self, policy_path):
        """
        Applies a Stack Policy.

        :param policy_path: The relative path of JSON file containing\
                the AWS Policy to apply.
        :type policy_path: str
        """
        with open(policy_path) as f:
            policy = f.read()

        self.logger.debug("%s - Setting Stack policy: \n%s", self.stack.name, policy)

        self.connection_manager.call(
            service="cloudformation",
            command="set_stack_policy",
            kwargs={"StackName": self.stack.external_name, "StackPolicyBody": policy},
        )
        self.logger.info("%s - Successfully set Stack Policy", self.stack.name)

    def get_policy(self):
        """
        Returns a Stack's Policy.

        :returns: The Stack's Stack Policy.
        :rtype: str
        """
        self.logger.debug("%s - Getting Stack Policy", self.stack.name)
        response = self.connection_manager.call(
            service="cloudformation",
            command="get_stack_policy",
            kwargs={"StackName": self.stack.external_name},
        )
        json_formatting = json.loads(
            response.get("StackPolicyBody", json.dumps("No Policy Information"))
        )
        return {self.stack.name: json_formatting}

    @add_stack_hooks
    def create_change_set(self, change_set_name):
        """
        Creates a Change Set with the name ``change_set_name``.

        :param change_set_name: The name of the Change Set.
        :type change_set_name: str
        """
        try:
            existing_status = self._get_status()
        except StackDoesNotExistError:
            existing_status = "PENDING"

        self.logger.info(
            "%s - Stack is in the %s state", self.stack.name, existing_status
        )

        change_set_type = (
            "CREATE"
            if existing_status in ["PENDING", "REVIEW_IN_PROGRESS"]
            else "UPDATE"
        )

        create_change_set_kwargs = {
            "StackName": self.stack.external_name,
            "Parameters": self._format_parameters(self.stack.parameters),
            "Capabilities": [
                "CAPABILITY_IAM",
                "CAPABILITY_NAMED_IAM",
                "CAPABILITY_AUTO_EXPAND",
            ],
            "ChangeSetName": change_set_name,
            "ChangeSetType": change_set_type,
            "NotificationARNs": self.stack.notifications,
            "Tags": [
                {"Key": str(k), "Value": str(v)} for k, v in self.stack.tags.items()
            ],
        }

        create_change_set_kwargs.update(self.stack.template.get_boto_call_parameter())
        create_change_set_kwargs.update(self._get_role_arn())

        try:
            self._create_change_set(change_set_name, create_change_set_kwargs)
        except Exception as err:
            self.logger.info(
                "%s - Failed creating Change Set '%s'\n%s",
                self.stack.name,
                change_set_name,
                err,
            )

    def _create_change_set(self, change_set_name, create_change_set_kwargs):
        self.logger.debug(
            "%s - Creating Change Set '%s'", self.stack.name, change_set_name
        )
        self.connection_manager.call(
            service="cloudformation",
            command="create_change_set",
            kwargs=create_change_set_kwargs,
        )
        # After the call successfully completes, AWS CloudFormation
        # starts creating the Change Set.
        self.logger.info(
            "%s - Successfully initiated creation of Change Set '%s'",
            self.stack.name,
            change_set_name,
        )

    def delete_change_set(self, change_set_name):
        """
        Deletes the Change Set ``change_set_name``.

        :param change_set_name: The name of the Change Set.
        :type change_set_name: str
        """
        # If the call successfully completes, AWS CloudFormation
        # successfully deleted the Change Set.
        try:
            self._delete_change_set(change_set_name)
            self.logger.info(
                "%s - Successfully deleted Change Set '%s'",
                self.stack.name,
                change_set_name,
            )
        except Exception as err:
            self.logger.info(
                "%s - Failed deleting Change Set '%s'\n%s",
                self.stack.name,
                change_set_name,
                err,
            )

    def _delete_change_set(self, change_set_name):
        self.logger.debug(
            "%s - Deleting Change Set '%s'", self.stack.name, change_set_name
        )
        self.connection_manager.call(
            service="cloudformation",
            command="delete_change_set",
            kwargs={
                "ChangeSetName": change_set_name,
                "StackName": self.stack.external_name,
            },
        )

    def describe_change_set(self, change_set_name):
        """
        Describes the Change Set ``change_set_name``.

        :param change_set_name: The name of the Change Set.
        :type change_set_name: str
        :returns: The description of the Change Set.
        :rtype: dict
        """
        return_val = {}

        try:
            return_val = self._describe_change_set(change_set_name)
        except Exception as err:
            self.logger.info(
                "%s - Failed describing Change Set '%s'\n%s",
                self.stack.name,
                change_set_name,
                err,
            )

        return return_val

    def _describe_change_set(self, change_set_name):
        self.logger.debug(
            "%s - Describing Change Set '%s'", self.stack.name, change_set_name
        )
        return self.connection_manager.call(
            service="cloudformation",
            command="describe_change_set",
            kwargs={
                "ChangeSetName": change_set_name,
                "StackName": self.stack.external_name,
            },
        )

    def execute_change_set(self, change_set_name):
        """
        Executes the Change Set ``change_set_name``.

        :param change_set_name: The name of the Change Set.
        :type change_set_name: str
        :returns: The Stack status
        :rtype: str
        """
        self._protect_execution()
        change_set = self.describe_change_set(change_set_name)
        status = change_set.get("Status")
        reason = change_set.get("StatusReason")

        return_val = 0

        if status == "FAILED" and self._change_set_creation_failed_due_to_no_changes(
            reason
        ):
            self.logger.info(
                "Skipping ChangeSet on Stack: {} - there are no changes".format(
                    change_set.get("StackName")
                )
            )
            return return_val

        try:
            return_val = self._execute_change_set(change_set_name)
        except Exception as err:
            self.logger.info(
                "%s - Failed describing Change Set '%s'\n%s",
                self.stack.name,
                change_set_name,
                err,
            )

        return return_val

    def _execute_change_set(self, change_set_name):
        self.logger.debug(
            "%s - Executing Change Set '%s'", self.stack.name, change_set_name
        )
        response = self.connection_manager.call(
            service="cloudformation",
            command="execute_change_set",
            kwargs={
                "ChangeSetName": change_set_name,
                "StackName": self.stack.external_name,
            },
        )
        status = self._wait_for_completion(boto_response=response)
        return status

    def _change_set_creation_failed_due_to_no_changes(self, reason: str) -> bool:
        """
        Indicates the change set failed when it was created because there were actually
        no changes introduced from the change set.

        :param reason: The reason reported by CloudFormation for the Change Set failure
        """
        reason = reason.lower()
        no_change_substrings = (
            "submitted information didn't contain changes",
            "no updates are to be performed",  # The reason returned for SAM templates
        )

        for substring in no_change_substrings:
            if substring in reason:
                return True
        return False

    def list_change_sets(self, url=False):
        """
        Lists the Stack's Change Sets.

        :param url: Write out a console URL instead.
        :type url: bool

        :returns: The Stack's Change Sets.
        :rtype: dict or list
        """
        response = self._list_change_sets()
        summaries = response.get("Summaries", [])

        if url:
            summaries = self._convert_to_url(summaries)

        return {self.stack.name: summaries}

    def _list_change_sets(self):
        self.logger.debug("%s - Listing change sets", self.stack.name)
        try:
            return self.connection_manager.call(
                service="cloudformation",
                command="list_change_sets",
                kwargs={"StackName": self.stack.external_name},
            )
        except botocore.exceptions.ClientError:
            return []

    def _convert_to_url(self, summaries):
        """
        Convert the list_change_sets response from
        CloudFormation to a URL in the AWS Console.
        """
        new_summaries = []

        for summary in summaries:
            stack_id = summary["StackId"]
            change_set_id = summary["ChangeSetId"]

            region = self.stack.region
            encoded = urllib.parse.urlencode(
                {"stackId": stack_id, "changeSetId": change_set_id}
            )

            new_summaries.append(
                f"https://{region}.console.aws.amazon.com/cloudformation/home?"
                f"region={region}#/stacks/changesets/changes?{encoded}"
            )

        return new_summaries

    def generate(self):
        """
        Returns the Template for the Stack. An alias for
        dump_template for historical reasons.
        """
        return self.dump_template()

    @add_stack_hooks
    def validate(self):
        """
        Validates the Stack's CloudFormation Template.

        Raises an error if the Template is invalid.

        :returns: Validation information about the Template.
        :rtype: dict
        :raises: botocore.exceptions.ClientError
        """
        self.logger.debug("%s - Validating Template", self.stack.name)
        response = self.connection_manager.call(
            service="cloudformation",
            command="validate_template",
            kwargs=self.stack.template.get_boto_call_parameter(),
        )
        self.logger.debug(
            "%s - Validate Template response: %s", self.stack.name, response
        )
        return response

    def estimate_cost(self):
        """
        Estimates a Stack's cost.

        :returns: An estimate of the Stack's cost.
        :rtype: dict
        :raises: botocore.exceptions.ClientError
        """
        self.logger.debug("%s - Estimating template cost", self.stack.name)

        parameters = [
            {"ParameterKey": key, "ParameterValue": value}
            for key, value in self.stack.parameters.items()
        ]

        kwargs = self.stack.template.get_boto_call_parameter()
        kwargs.update({"Parameters": parameters})

        response = self.connection_manager.call(
            service="cloudformation", command="estimate_template_cost", kwargs=kwargs
        )
        self.logger.debug(
            "%s - Estimate Stack cost response: %s", self.stack.name, response
        )
        return response

    def get_status(self):
        """
        Returns the Stack's status.

        :returns: The Stack's status.
        :rtype: sceptre.stack_status.StackStatus
        """
        try:
            return self._get_status()
        except StackDoesNotExistError:
            return "PENDING"

    def _format_parameters(self, parameters):
        """
        Converts CloudFormation parameters to the format used by Boto3.

        :param parameters: A dictionary of parameters.
        :type parameters: dict
        :returns: A list of the formatted parameters.
        :rtype: list
        """
        formatted_parameters = []
        for name, value in parameters.items():
            if value is None:
                continue
            if isinstance(value, list):
                value = ",".join(value)
            formatted_parameters.append({"ParameterKey": name, "ParameterValue": value})

        return formatted_parameters

    def _get_role_arn(self):
        """
        Returns the Role ARN assumed by CloudFormation when building a Stack.

        Returns an empty dict if no Role is to be assumed.

        :returns: The a Role ARN
        :rtype: dict
        """
        if self.stack.cloudformation_service_role:
            return {"RoleARN": self.stack.cloudformation_service_role}
        else:
            return {}

    def _get_stack_timeout(self):
        """
        Return the timeout before considering the Stack to be failing.

        Returns an empty dict if no timeout is set.
        :returns: the creation/update timeout
        :rtype: dict
        """
        if self.stack.stack_timeout:
            return {"TimeoutInMinutes": self.stack.stack_timeout}
        else:
            return {}

    def _protect_execution(self):
        """
        Raises a ProtectedStackError if protect == True.

        :raises: sceptre.exceptions.ProtectedStackError
        """
        if self.stack.protected:
            raise ProtectedStackError(
                "Cannot perform action on '{0}': Stack protection is "
                "currently enabled".format(self.stack.name)
            )

    def _wait_for_completion(
        self, timeout=0, boto_response: Optional[dict] = None
    ) -> StackStatus:
        """
        Waits for a Stack operation to finish. Prints CloudFormation events
        while it waits.

        :param timeout: Timeout before returning, in minutes.
        :param boto_response: Response from the boto call which initiated the stack change.

        :returns: The final Stack status.
        """
        timeout = 60 * timeout

        def timed_out(elapsed):
            return elapsed >= timeout if timeout else False

        status = StackStatus.IN_PROGRESS

        most_recent_event_datetime = extract_datetime_from_aws_response_headers(
            boto_response
        ) or (datetime.now(tzutc()) - timedelta(seconds=3))

        elapsed = 0
        while status == StackStatus.IN_PROGRESS and not timed_out(elapsed):
            status = self._get_simplified_status(self._get_status())
            most_recent_event_datetime = self._log_new_events(
                most_recent_event_datetime
            )
            time.sleep(4)
            elapsed += 4

        return status

    def _get_status(self):
        try:
            status = self.describe()["Stacks"][0]["StackStatus"]
        except botocore.exceptions.ClientError as exp:
            if exp.response["Error"]["Message"].endswith("does not exist"):
                raise StackDoesNotExistError(exp.response["Error"]["Message"])
            else:
                raise exp
        return status

    @staticmethod
    def _get_simplified_status(status):
        """
        Returns the simplified Stack Status.

        The simplified Stack status is represented by the struct
        ``sceptre.StackStatus()`` and can take one of the following options:

        * complete
        * in_progress
        * failed

        :param status: The CloudFormation Stack status to simplify.
        :type status: str
        :returns: The Stack's simplified status
        :rtype: sceptre.stack_status.StackStatus
        """
        if status.endswith("ROLLBACK_COMPLETE"):
            return StackStatus.FAILED
        elif status.endswith("_COMPLETE"):
            return StackStatus.COMPLETE
        elif status.endswith("_IN_PROGRESS"):
            return StackStatus.IN_PROGRESS
        elif status.endswith("_FAILED"):
            return StackStatus.FAILED
        else:
            raise UnknownStackStatusError("{0} is unknown".format(status))

    def _log_new_events(self, after_datetime: datetime) -> datetime:
        """
        Log the latest Stack events while the Stack is being built.

        :param after_datetime: Only events after this datetime will be logged.
        :returns: The datetime of the last logged event or after_datetime if no events were logged.
        """
        events = self.describe_events()["StackEvents"]
        events.reverse()
        new_events = [event for event in events if event["Timestamp"] > after_datetime]
        for event in new_events:
            stack_event_status = [
                self.stack.name,
                event["LogicalResourceId"],
                event["ResourceType"],
                event["ResourceStatus"],
                event.get("ResourceStatusReason", ""),
            ]
            if "HookStatus" in event:
                stack_event_status.extend(
                    [
                        event["HookType"],
                        event["HookStatus"],
                        event.get("HookStatusReason", ""),
                        event["HookFailureMode"],
                    ]
                )
            self.logger.info(" ".join(stack_event_status))
            after_datetime = event["Timestamp"]
        return after_datetime

    def wait_for_cs_completion(self, change_set_name):
        """
        Waits while the Stack Change Set status is "pending".

        :param change_set_name: The name of the Change Set.
        :type change_set_name: str
        :returns: The Change Set's status.
        :rtype: sceptre.stack_status.StackChangeSetStatus
        """
        while True:
            status = self._get_cs_status(change_set_name)
            if status != StackChangeSetStatus.PENDING:
                break
            time.sleep(2)

        return status

    def _get_cs_status(self, change_set_name):
        """
        Returns the status of a Change Set.

        :param change_set_name: The name of the Change Set.
        :type change_set_name: str
        :returns: The Change Set's status.
        :rtype: sceptre.stack_status.StackChangeSetStatus
        """
        cs_description = self.describe_change_set(change_set_name)

        cs_status = cs_description["Status"]
        cs_reason = cs_description.get("StatusReason")
        cs_exec_status = cs_description["ExecutionStatus"]
        possible_statuses = [
            "CREATE_PENDING",
            "CREATE_IN_PROGRESS",
            "CREATE_COMPLETE",
            "DELETE_COMPLETE",
            "FAILED",
        ]
        possible_execution_statuses = [
            "UNAVAILABLE",
            "AVAILABLE",
            "EXECUTE_IN_PROGRESS",
            "EXECUTE_COMPLETE",
            "EXECUTE_FAILED",
            "OBSOLETE",
        ]

        if cs_status not in possible_statuses:
            raise UnknownStackChangeSetStatusError(
                "Status {0} is unknown".format(cs_status)
            )
        if cs_exec_status not in possible_execution_statuses:
            raise UnknownStackChangeSetStatusError(
                "ExecutionStatus {0} is unknown".format(cs_status)
            )

        if cs_status == "CREATE_COMPLETE" and cs_exec_status == "AVAILABLE":
            return StackChangeSetStatus.READY
        elif cs_status in [
            "CREATE_PENDING",
            "CREATE_IN_PROGRESS",
            "CREATE_COMPLETE",
        ] and cs_exec_status in ["UNAVAILABLE", "AVAILABLE"]:
            return StackChangeSetStatus.PENDING
        elif (
            cs_status == "FAILED"
            and cs_reason is not None
            and self._change_set_creation_failed_due_to_no_changes(cs_reason)
        ):
            return StackChangeSetStatus.NO_CHANGES
        elif cs_status in ["DELETE_COMPLETE", "FAILED"] or cs_exec_status in [
            "EXECUTE_IN_PROGRESS",
            "EXECUTE_COMPLETE",
            "EXECUTE_FAILED",
            "OBSOLETE",
        ]:
            return StackChangeSetStatus.DEFUNCT
        else:  # pragma: no cover
            raise Exception("This else should not be reachable.")

    def fetch_remote_template(self) -> Optional[str]:
        """
        Returns the Template for the remote Stack

        :returns: the template body.
        """
        self.logger.debug(f"{self.stack.name} - Fetching remote template")

        original_template = self._fetch_original_template_stage()

        if isinstance(original_template, dict):
            # While not documented behavior, boto3 will attempt to deserialize the TemplateBody
            # with json.loads and return the template as a dict if it is successful; otherwise (such
            # as in when the template is in yaml, it will return the string. Therefore, we need to
            # dump the template to json if we get a dict.
            original_template = json.dumps(original_template, indent=4)

        return original_template

    def _fetch_original_template_stage(self) -> Optional[Union[str, dict]]:
        try:
            response = self.connection_manager.call(
                service="cloudformation",
                command="get_template",
                kwargs={
                    "StackName": self.stack.external_name,
                    "TemplateStage": "Original",
                },
            )
            return response["TemplateBody"]
            # Sometimes boto returns a string, sometimes a dictionary
        except botocore.exceptions.ClientError as e:
            # AWS returns a ValidationError if the stack doesn't exist
            if e.response["Error"]["Code"] == "ValidationError":
                return None
            raise

    def fetch_remote_template_summary(self):
        return self._get_template_summary(StackName=self.stack.external_name)

    def fetch_local_template_summary(self):
        boto_call_parameter = self.stack.template.get_boto_call_parameter()
        return self._get_template_summary(**boto_call_parameter)

    def _get_template_summary(self, **kwargs) -> Optional[dict]:
        try:
            template_summary = self.connection_manager.call(
                service="cloudformation", command="get_template_summary", kwargs=kwargs
            )
            return template_summary
        except botocore.exceptions.ClientError as e:
            error_response = e.response["Error"]
            if (
                error_response["Code"] == "ValidationError"
                and "does not exist" in error_response["Message"]
            ):
                return None
            raise

    @add_stack_hooks
    def diff(self, stack_differ: "StackDiffer") -> "StackDiff":
        """
        Returns a diff of local and deployed template and stack configuration using a specific diff
        library.

        :param stack_differ: The differ to use
        :returns: A StackDiff object with the full, computed diff
        """
        return stack_differ.diff(self)

    @add_stack_hooks
    def drift_detect(self) -> Dict[str, str]:
        """
        Show stack drift for a running stack.

        :returns: The stack drift detection status.
            If the stack does not exist, we return a detection and
            stack drift status of STACK_DOES_NOT_EXIST.
            If drift detection times out after 5 minutes, we return
            TIMED_OUT.
        """
        try:
            self._get_status()
        except StackDoesNotExistError:
            self.logger.info(f"{self.stack.name} - Does not exist.")
            return {
                "DetectionStatus": "STACK_DOES_NOT_EXIST",
                "StackDriftStatus": "STACK_DOES_NOT_EXIST",
            }

        response = self._detect_stack_drift()
        detection_id = response["StackDriftDetectionId"]

        try:
            response = self._wait_for_drift_status(detection_id)
        except TimeoutError as exc:
            self.logger.info(f"{self.stack.name} - {exc}")
            response = {"DetectionStatus": "TIMED_OUT", "StackDriftStatus": "TIMED_OUT"}

        return response

    @add_stack_hooks
    def drift_show(self, drifted: bool = False) -> Tuple[str, dict]:
        """
        Detect drift status on stacks.

        :param drifted: Filter out IN_SYNC resources.
        :returns: The detection status and resource drifts.
        """
        response = self.drift_detect()
        detection_status = response["DetectionStatus"]

        if detection_status in ["DETECTION_COMPLETE", "DETECTION_FAILED"]:
            response = self._describe_stack_resource_drifts()
        elif detection_status in ["TIMED_OUT", "STACK_DOES_NOT_EXIST"]:
            response = {"StackResourceDriftStatus": detection_status}
        else:
            raise Exception("Not expected to be reachable")

        response = self._filter_drifts(response, drifted)
        return (detection_status, response)

    def _wait_for_drift_status(self, detection_id: str) -> dict:
        """
        Waits for drift detection to complete.

        :param detection_id: The drift detection ID.
        :returns: The response from describe_stack_drift_detection_status.
        """
        timeout = 300
        sleep_interval = 10
        elapsed = 0

        while True:
            if elapsed >= timeout:
                raise TimeoutError(f"Timed out after {elapsed} seconds")

            self.logger.info(f"{self.stack.name} - Waiting for drift detection")
            response = self._describe_stack_drift_detection_status(detection_id)
            detection_status = response["DetectionStatus"]

            self._log_drift_status(response)

            if detection_status == "DETECTION_IN_PROGRESS":
                time.sleep(sleep_interval)
                elapsed += sleep_interval
            else:
                return response

    def _log_drift_status(self, response: dict) -> None:
        """
        Log the drift status while waiting for
        drift detection to complete.
        """
        keys = [
            "StackDriftDetectionId",
            "DetectionStatus",
            "DetectionStatusReason",
            "StackDriftStatus",
        ]

        for key in keys:
            if key in response:
                self.logger.debug(f"{self.stack.name} - {key} - {response[key]}")

    def _detect_stack_drift(self) -> dict:
        self.logger.info(f"{self.stack.name} - Detecting Stack Drift")

        return self.connection_manager.call(
            service="cloudformation",
            command="detect_stack_drift",
            kwargs={"StackName": self.stack.external_name},
        )

    def _describe_stack_drift_detection_status(self, detection_id: str) -> dict:
        self.logger.info(f"{self.stack.name} - Describing Stack Drift Detection Status")

        return self.connection_manager.call(
            service="cloudformation",
            command="describe_stack_drift_detection_status",
            kwargs={"StackDriftDetectionId": detection_id},
        )

    def _describe_stack_resource_drifts(self) -> dict:
        """
        Detects stack resource_drifts for a running stack.
        """
        self.logger.info(f"{self.stack.name} - Describing Stack Resource Drifts")

        return self.connection_manager.call(
            service="cloudformation",
            command="describe_stack_resource_drifts",
            kwargs={"StackName": self.stack.external_name},
        )

    def _filter_drifts(self, response: dict, drifted: bool) -> dict:
        """
        The filtered response after filtering out StackResourceDriftStatus.
        :param drifted: Filter out IN_SYNC resources from CLI --drifted.
        """
        if "StackResourceDrifts" not in response:
            return response

        result = {"StackResourceDrifts": []}
        include_all_drift_statuses = not drifted

        for drift in response["StackResourceDrifts"]:
            is_drifted = drift["StackResourceDriftStatus"] != "IN_SYNC"
            if include_all_drift_statuses or is_drifted:
                result["StackResourceDrifts"].append(drift)

        return result

    @add_stack_hooks
    def dump_config(self):
        """
        Dump the config for a stack.
        """
        return self.stack.config

    @add_stack_hooks_with_aliases([generate.__name__])
    def dump_template(self):
        """
        Dump the template for the Stack. An alias for generate
        for historical reasons.
        """
        return self.stack.template.body



# -*- coding: utf-8 -*-

import logging


__author__ = "Cloudreach"
__email__ = "sceptre@cloudreach.com"


# Set up logging to ``/dev/null`` like a library is supposed to.
# http://docs.python.org/3.3/howto/logging.html#configuring-logging-for-a-library
class NullHandler(logging.Handler):  # pragma: no cover
    def emit(self, record):
        pass


logging.getLogger("sceptre").addHandler(NullHandler())
